#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 28 11:41:23 2020

@author: diego
"""

#La información de este tutorial está basada en: https://docs.scipy.org/doc/scipy/reference/tutorial/general.html

# Se importan las librerías básicas para manejo de arrays y generación de gráficos

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

# subpaquetes de la librería SciPy

#Subpackage      Description
#
#cluster         Clustering algorithms
#constants       Physical and mathematical constants
#fftpack         Fast Fourier Transform routines
#integrate       Integration and ordinary differential equation solvers
#interpolate     Interpolation and smoothing splines
#io              Input and Output
#linalg          Linear algebra
#ndimage         N-dimensional image processing
#odr             Orthogonal distance regression
#optimize        Optimization and root-finding routines
#signal          Signal processing
#sparse          Sparse matrices and associated routines
#spatial         Spatial data structures and algorithms
#special         Special functions
#stats           Statistical distributions and functions

# Tema 1 - interpolación de datos +++++++++++++++++++++++++++++++++++++++++++++

# un ejemplo de datos experimentales extraído del Cap. 3 de Scientific Computing with Matlab and Octave.  A. Quarteroni, F. Saleri.
x_exp = np.arange(-55, 66, 10)
y_exp = np.array([-3.25, -3.37, -3.35, -3.2, -3.12, -3.02, -3.02, -3.07, -3.17, -3.32, -3.3, -3.22, -3.1])
n_exp = x_exp.size

# graficamos los resultados experimentales
plt.plot(x_exp,y_exp,'o')
plt.grid()
plt.xlabel('Latitud')
plt.ylabel('Concentración')
plt.xlim(-65, 75)
plt.ylim(-3.4, -3)
plt.show()

# con n_exp puntos, se puede ajustar un polinomio de orden n_exp-1
p = np.polyfit(x_exp,y_exp,n_exp-1)
x_fit = np.linspace(-60,70,200)             # genero el nuevo rango de x
y_fit = np.polyval(p,x_fit)

# graficamos los resultados experimentales y el polinomio de Lagrange de orden n_exp-1
plt.plot(x_exp,y_exp,'o',x_fit,y_fit,'-')
plt.grid()
plt.xlabel('Latitud')
plt.ylabel('Concentración')
plt.xlim(-65, 75)
plt.ylim(-3.4, -3)
plt.show()

# conclusión, no conviene generar polinomios interpolantes de alto orden.  En los extremos el comportamiento no es representativo.

# probamos algo de menor orden...
p4 = np.polyfit(x_exp,y_exp,4)
y_fit4 = np.polyval(p4,x_fit)

# graficamos los resultados experimentales y el polinomio de Lagrange de orden n_exp-1
plt.plot(x_exp,y_exp,'o', x_fit,y_fit,'-',x_fit,y_fit4,'--')
plt.grid()
plt.xlabel('Latitud')
plt.ylabel('Concentración')
plt.xlim(-65, 75)
plt.ylim(-3.4, -3)
plt.legend(['Exp Data','Poly 12','Poly 4'])
plt.show()

# probamos algo de menor orden...
p6 = np.polyfit(x_exp,y_exp,6)
y_fit6 = np.polyval(p6,x_fit)

# graficamos los resultados experimentales y el polinomio de Lagrange de orden n_exp-1
plt.plot(x_exp,y_exp,'o', x_fit,y_fit,'-',x_fit,y_fit4,'--',x_fit,y_fit6,'r-.')
plt.grid()
plt.xlabel('Latitud')
plt.ylabel('Concentración')
plt.xlim(-65, 75)
plt.ylim(-3.4, -3)
plt.legend(['Exp Data','Poly 12','Poly 4','Poly 6'])
plt.show()

# Otra forma de realizarlo, esta vez usando SciPy: https://docs.scipy.org/doc/scipy/reference/tutorial/interpolate.html

from scipy.interpolate import interp1d

y_lin = interp1d(x_exp,y_exp)       # interpolación lineal a trozos.  NOTAR LA DIFERENCIA CON GRAFICAR CON LINEAS
y_sp3 = interp1d(x_exp,y_exp, kind = 'cubic')       # interpolación con un spline cúbico

x_fit2 = np.linspace(-55,65,200)

plt.plot(x_exp,y_exp,'o', x_fit2,y_lin(x_fit2),'-',x_fit2,y_sp3(x_fit2),'--')
plt.grid()
plt.xlabel('Latitud')
plt.ylabel('Concentración')
plt.xlim(-65, 75)
plt.ylim(-3.4, -3)
plt.legend(['Exp Data','Lineal','Spline 3'], loc='upper left')
plt.show()

# finalmente, comparamos los polimonios con el spline

plt.plot(x_exp,y_exp,'o', x_fit,y_fit,'-',x_fit,y_fit4,'--',x_fit2,y_sp3(x_fit2),'-.')
plt.grid()
plt.xlabel('Latitud')
plt.ylabel('Concentración')
plt.xlim(-65, 75)
plt.ylim(-3.4, -3)
plt.legend(['Exp Data','Poly 12','Poly 4','Spline 3'], loc='upper left')
plt.show()

# mencionar la imposibilidad/inconveniente de extrapolar.

# Ajuste lineal +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

x = np.arange(0,16,1,dtype='float64')
y1 = x*1.37 + 2.33                                      # una recta conocida

rng = np.random.default_rng(0)
amplitud = 6
noise = rng.random(x.shape)*amplitud - amplitud/2         # ruido de media 0 y amplitud "amplitud"
y2 = y1 + noise

plt.plot(x,y1,'o',x,y2,'rs')
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['Datos Base','Datos con ruido'], loc='upper left')
plt.show()

print(np.polyfit(x,y1,1,full='True'))           # vemos que el comando polyfit con contiene el valor de R²

p = np.polyfit(x,y2,1)          # ajuste de los datos con ruido (puede dar distinto a 1.37 y 2.33)

x3 = np.linspace(0,15,100)
y3 = p[0]*x3 + p[1]

plt.plot(x,y1,'o',x,y2,'s',x3,y3,'r-')
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['Datos Base','Datos con ruido','Ajuste lineal'], loc='upper left')
plt.show()

from sklearn.metrics import r2_score            # se debe instalar la librería scikit-learn, ver: https://scikit-learn.org/stable/install.html

y2_Fit = p[0]*x + p[1]                          # valores de y2 de acuerdo al ajuste
r2 = r2_score(y2, y2_Fit)                       # se ingresa el valor de Y real (con ruido) y el de Y ajustado

# en vez de andar cargando una librería entera, los estudiantes podrían probar de hacerse su propia función de ajuste lineal que les devuelva los dos parámetros de la recta y el R²

# Por último, otras opciones del método interp1d ++++++++++++++++++++++++++++++

f1 = interp1d(x, y2, kind='nearest')
f2 = interp1d(x, y2, kind='previous')
f3 = interp1d(x, y2, kind='next')

plt.plot(x,y2,'s',x3,f1(x3),'-',x3,f2(x3),'--',x3,f3(x3),'-.')
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['Datos con ruido','nearest','previous','next'], loc='upper left')
plt.show()

# Derivadas +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#def func(x):
#    y = np.empty(x.shape)
#    for i in range(x.size):
#        y[i] = math.sin(2*math.pi*x[i]**0.5)*x[i]             # para ver el uso de funciones trigonométricas, ver: https://docs.python.org/3/library/math.html
#    return y

def funcion(x):                                    # defino la función que quiero derivar (tiene que manejarse con un escalar como argumento, no vale un array) 
    return np.sin(2*np.pi*x**0.5)*x

x4 = np.linspace(0.001,4,100)

fun = np.empty(x4.shape)        
for i in range(x4.size):
    fun[i] = funcion(x4[i])                        # la evalúo en todo el vector x3

plt.plot(x4,fun,'-')                            # la grafico
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.ylim(-6,7)
plt.legend("f(x)", loc='upper left')
plt.show()

from scipy.misc import derivative                # importo el método de derivación centrada "derivative". funciona solo para argumentos escalares.

dfun = np.empty(x4.shape)
for i in range(x4.size):                        # derivo la función en todo el dominio (es válido para escalares)
    dfun[i] = derivative(funcion,x4[i],dx=1e-6)     

plt.plot(x4,fun,'-',x4,dfun,'-.')                            # la grafico
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.ylim(-6,7)
plt.legend(["f(x)","f'(x)"], loc='upper left')
plt.show()

# tenemos otra forma de derivar con valores centrados, si necesidad de un método.  Lo más general se puede leer acá, aunque tiene algunos errores: https://www.math.ubc.ca/~pwalls/math-python/differentiation/differentiation/

def DeriCentrada(x,y):              # funcion que deriva en todo el dominio, proporcional a h²
    dy = np.empty(x.shape)
    for i in range(x.size):
        if i==0:
            dy[i] = (-3*y[i]+4*y[i+1]-y[i+2])/(x[i+2]-x[i])
        elif 0<i<x.size-1:
            dy[i] = (y[i+1]-y[i-1])/(x[i+1]-x[i-1])
        else:
            dy[i] = (y[i-2]-4*y[i-1]+3*y[i])/(x[i]-x[i-2])      
    return dy
    
dfun2 = DeriCentrada(x4,fun)

plt.plot(x4,dfun,'b-',x4,dfun2,'y--')                            # veo que dan parecidas
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.ylim(-6,7)
plt.legend(["f'(x) Scipy","f'(x) Centrada"], loc='upper left')
plt.show()

plt.plot(x4,dfun-dfun2,'-')                            # veo que dan parecidas
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
#plt.ylim(-6,7)
#plt.legend(["f'(x) Scipy","f'(x) Centrada"], loc='upper left')
plt.show()


# Integrales ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import scipy.integrate as integrate

#help(integrate)
# Methods for Integrating Functions given function object.
#
#   quad          -- General purpose integration.
#   dblquad       -- General purpose double integration.
#   tplquad       -- General purpose triple integration.
#   fixed_quad    -- Integrate func(x) using Gaussian quadrature of order n.
#   quadrature    -- Integrate with given tolerance using Gaussian quadrature.
#   romberg       -- Integrate func using Romberg integration.
#
# Methods for Integrating Functions given fixed samples.
#
#   trapz         -- Use trapezoidal rule to compute integral from samples.
#   cumtrapz      -- Use trapezoidal rule to cumulatively compute integral.
#   simps         -- Use Simpson's rule to compute integral from samples.
#   romb          -- Use Romberg Integration to compute integral from
#                    (2**k + 1) evenly-spaced samples.

from scipy.integrate import quad

(Iquad,err) = quad(funcion,x4[0],x4[-1])          # esto nos da la integral definida entre dos límites, incluido su error

Intfun = np.empty(x4.shape)
for i in range(x4.size):                        # Función integral a lo largo del dominio
    (Intfun[i],err)= quad(funcion,x4[0],x4[i])
    
plt.plot(x4,fun,'-',x4,dfun,'-.',x4,Intfun,'--')                            # la grafico
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.ylim(-6,7)
plt.legend(["f(x)","f'(x)","F(x)"], loc='upper left')
plt.show()

# También podemos integrar dos arrays X e Y

from scipy.integrate import simps,trapz              # regla de Simpson de integración

ISimp = simps(fun, x4)

IntfunS = np.empty(x4.shape)
for i in range(x4.size):                        # Función integral a lo largo del dominio
    if i==0:
        IntfunS[0] = 0
    elif i==1:
        IntfunS[1] = trapz(fun[0:i],x4[0:i])
    else:
        IntfunS[i] = simps(fun[0:i],x4[0:i])
        
plt.plot(x4,Intfun,'-',x4,IntfunS,'--')                            # la grafico
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
#plt.ylim(-6,7)
plt.legend(["Quad","Simpson"], loc='upper left')
plt.show()